from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from rasa_core.utils import EndpointConfig
import json
import logging
import requests
from rasa_core import constants
logger = logging.getLogger(__name__)

INTENT_MESSAGE_PREFIX = "/"

from rasa_core.interpreter import NaturalLanguageInterpreter, RasaNLUHttpInterpreter

class RasaMultiNLUHttpInterpreter(RasaNLUHttpInterpreter):
    def __init__(self, models=None, endpoint=None, project_name='default'):
        # type: (Text, EndpointConfig, Text) -> None

        self.models = models
        self.project_name = project_name

        if endpoint:
            self.endpoint = endpoint
        else:
            self.endpoint = EndpointConfig(constants.DEFAULT_SERVER_URL)

    def get_model(self, lang):
        return self.models.get(lang)

    def parse(self, text, lang):
        """Parse a text message.

        Return a default value if the parsing of the text failed."""

        default_return = {"intent": {"name": "", "confidence": 0.0},
                          "entities": [], "text": ""}
        result = self._rasa_http_parse(text, lang)

        return result if result is not None else default_return

    def _rasa_http_parse(self, text, lang):
        """Send a text message to a running rasa NLU http server.

        Return `None` on failure."""

        if not self.endpoint:
            logger.error(
                    "Failed to parse text '{}' using rasa NLU over http. "
                    "No rasa NLU server specified!".format(text))
            return None

        params = {
            "token": self.endpoint.token,
            "model": self.get_model(lang),
            "project": self.project_name,
            "q": text
        }
        url = "{}/parse".format(self.endpoint.url)
        try:
            result = requests.get(url, params=params)
            if result.status_code == 200:
                return result.json()
            else:
                logger.error(
                        "Failed to parse text '{}' using rasa NLU over http. "
                        "Error: {}".format(text, result.text))
                return None
        except Exception as e:
            logger.error(
                    "Failed to parse text '{}' using rasa NLU over http. "
                    "Error: {}".format(text, e))
            return None



import os
import logging
import warnings
import time

from rasa_core.agent import Agent, start_model_pulling_in_worker, _update_model_from_server
from rasa_core.domain import TemplateDomain
from rasa_core.interpreter import (
    RegexInterpreter,
    NaturalLanguageInterpreter,
    INTENT_MESSAGE_PREFIX)
from rasa_core.nlg import NaturalLanguageGenerator
from rasa_core.policies.ensemble import PolicyEnsemble
from rasa_core.events import UserUttered
from rasa_core.processor import MessageProcessor
from rasa_core.dispatcher import Dispatcher

from rasa_addons.multilingual_interpreter import RasaMultiNLUHttpInterpreter
from rasa_addons.rules import Rules
from rasa_core.utils import EndpointConfig
from threading import Thread

logging.basicConfig()
logger = logging.getLogger()


class SuperAgent(Agent):
    def __init__(
            self,
            domain,  # type: Union[Text, Domain]
            policies=None,  # type: Optional[Union[PolicyEnsemble, List[Policy]]
            interpreter=None,  # type: Optional[NaturalLanguageInterpreter]
            generator=None,  # type: Union[EndpointConfig, NLG]
            tracker_store=None,  # type: Optional[TrackerStore]
            action_endpoint=None,  # type: Optional[EndpointConfig]
            fingerprint=None,  # type: Optional[Text]
            create_dispatcher=None,  # type: Optional[LambdaType]
            create_nlg=None,  # type: Optional[LambdaType],
            rules=None,  # type: Optional[EndpointConfig, str]
    ):
        self.processor = None
        self.create_dispatcher = create_dispatcher
        self.create_nlg = create_nlg

        self.rules = None
        # Initializing variables with the passed parameters.
        self.domain = self._create_domain(domain)
        self.policy_ensemble = self._create_ensemble(policies)

        if not isinstance(interpreter, NaturalLanguageInterpreter):
            if interpreter is not None:
                logger.warning(
                    "Passing a value for interpreter to an agent "
                    "where the value is not an interpreter "
                    "is deprecated. Construct the interpreter, before"
                    "passing it to the agent, e.g. "
                    "`interpreter = NaturalLanguageInterpreter.create("
                    "nlu)`.")
            interpreter = NaturalLanguageInterpreter.create(interpreter, None)

        self.interpreter = interpreter

        self.nlg = NaturalLanguageGenerator.create(generator, self.domain) if self.create_nlg is None else self.create_nlg(generator, self.domain)
        self.tracker_store = self.create_tracker_store(
            tracker_store, self.domain)
        self.action_endpoint = action_endpoint

        self._set_fingerprint(fingerprint)

    @staticmethod
    def get_rules(rules_source):
        if isinstance(rules_source, EndpointConfig):
            return Rules.load_from_remote(rules_source)
        elif isinstance(rules_source, str):
            return Rules.load_from_file(rules_source)
        elif rules_source is not None:
            raise ValueError('Rules must be either a path to a yaml file, or an endpoint of which the GET method '
                             'returns rules in a JSON format')
        else:
            return None



    @classmethod
    def load(cls,
             path,
             domain=None,
             policies=None,
             interpreter=None,
             generator=None,
             tracker_store=None,
             action_endpoint=None,
             rules=None,
             create_dispatcher=None,
             model_server=None,  # type: Optional[EndpointConfig]
             wait_time_between_pulls=None,  # type: Optional[int]
             create_nlg=None):
        # type: (Text, Any, Optional[TrackerStore]) -> Agent

        if not path and not domain:
            raise ValueError("You need to provide a valid directory where "
                             "to load the agent from when calling "
                             "`Agent.load`.")

        if os.path.isfile(path):
            raise ValueError("You are trying to load a MODEL from a file "
                             "('{}'), which is not possible. \n"
                             "The persisted path should be a directory "
                             "containing the various model files. \n\n"
                             "If you want to load training data instead of "
                             "a model, use `agent.load_data(...)` "
                             "instead.".format(path))

        # We don't want to block if the path doesn't exist but a model_server is supplied
        if domain is None and model_server is None:
            domain = TemplateDomain.load(os.path.join(path, "domain.yml"))
        if policies is None and model_server is None:
            policies = PolicyEnsemble.load(path)

        # ensures the domain hasn't changed between test and train
        if model_server is None:
            domain.compare_with_specification(path)
        #
        # _interpreter = NaturalLanguageInterpreter.create(interpreter)
        # _tracker_store = cls.create_tracker_store(tracker_store, domain)

        agent = cls(
                domain=domain,
                policies=policies,
                interpreter=interpreter,
                tracker_store=tracker_store,
                generator=generator,
                action_endpoint=action_endpoint,
                rules=rules,
                create_dispatcher=create_dispatcher,
                create_nlg=create_nlg
        )
        if model_server:
            if wait_time_between_pulls:
                # continuously pull the model every `wait_time_between_pulls` seconds
                start_model_pulling_in_worker(model_server,
                                              wait_time_between_pulls,
                                              agent)
            else:
                # just pull the model once
                _update_model_from_server(model_server, agent)

        if rules:
            # Start worker if `wait_time_between_pulls` is set
            if isinstance(rules, EndpointConfig) and wait_time_between_pulls:
                # continuously pull the rules every `wait_time_between_pulls` seconds
                start_rules_pulling_in_worker(rules,
                                              wait_time_between_pulls,
                                              agent)
            # In all other cases we only want to load the rules once
            else:
                agent.rules = SuperAgent.get_rules(rules)
        return agent

    def create_processor(self, preprocessor=None):
        # type: (Callable[[Text], Text]) -> MessageProcessor
        """Instantiates a processor based on the set state of the agent."""

        self._ensure_agent_is_ready()
        self.processor = SuperMessageProcessor(self.interpreter,
                                               self.policy_ensemble,
                                               self.domain,
                                               self.tracker_store,
                                               self.nlg,
                                               message_preprocessor=preprocessor,
                                               action_endpoint=self.action_endpoint,
                                               create_dispatcher=self.create_dispatcher,
                                               rules=self.rules)
        return self.processor


class SuperMessageProcessor(MessageProcessor):
    def __init__(self,
                 interpreter,  # type: NaturalLanguageInterpreter
                 policy_ensemble,  # type: PolicyEnsemble
                 domain,  # type: Domain
                 tracker_store,  # type: TrackerStore
                 generator,  # type: NaturalLanguageGenerator
                 action_endpoint=None,  # type: Optional[EndpointConfig]
                 max_number_of_predictions=10,  # type: int
                 message_preprocessor=None,  # type: Optional[LambdaType]
                 on_circuit_break=None,  # type: Optional[LambdaType]
                 create_dispatcher=None,  # type: Optional[LambdaType]
                 rules=None  # type: Optional[Rules]
                 ):

        self.rules = rules
        super(SuperMessageProcessor, self).__init__(
            interpreter,
            policy_ensemble,
            domain,
            tracker_store,
            generator,
            action_endpoint,
            max_number_of_predictions,
            message_preprocessor,
            on_circuit_break
        )

        self.create_dispatcher = create_dispatcher
        if self.create_dispatcher is None:
            self.create_dispatcher = lambda sender_id, output_channel, nlg: Dispatcher(sender_id, output_channel, nlg)

    def _handle_message_with_tracker(self, message, tracker):
        # type: (UserMessage, DialogueStateTracker) -> None

        if message.parse_data:
            parse_data = message.parse_data
        else:
            parse_data = self._parse_message(message)

        # rules section #
        if self._rule_interrupts(parse_data, tracker, message):
            return
        # rules section - end #

        # don't ever directly mutate the tracker
        # - instead pass its events to log
        tracker.update(UserUttered(message.text, parse_data["intent"],
                                   parse_data["entities"], parse_data))
        # store all entities as slots
        for e in self.domain.slots_for_entities(parse_data["entities"]):
            tracker.update(e)

        logger.debug("Logged UserUtterance - "
                     "tracker now has {} events".format(len(tracker.events)))

    def _rule_interrupts(self, parse_data, tracker, message):
        if self.rules is not None:
            dispatcher = self.create_dispatcher(message.sender_id, message.output_channel, self.nlg)
            return self.rules.interrupts(dispatcher, parse_data, tracker, self._run_action)

    def _predict_and_execute_next_action(self, message, tracker):
        # this will actually send the response to the user

        dispatcher = self.create_dispatcher(message.sender_id,
                                            message.output_channel,
                                            self.nlg)
        # keep taking actions decided by the policy until it chooses to 'listen'
        should_predict_another_action = True
        num_predicted_actions = 0

        self._log_slots(tracker)

        # action loop. predicts actions until we hit action listen
        while (should_predict_another_action
               and self._should_handle_message(tracker)
               and num_predicted_actions < self.max_number_of_predictions):
            # this actually just calls the policy's method by the same name
            action, policy, confidence = self.predict_next_action(tracker)

            should_predict_another_action = self._run_action(action,
                                                             tracker,
                                                             dispatcher,
                                                             policy,
                                                             confidence)
            num_predicted_actions += 1

        if (num_predicted_actions == self.max_number_of_predictions and
                should_predict_another_action):
            # circuit breaker was tripped
            logger.warn(
                "Circuit breaker tripped. Stopped predicting "
                "more actions for sender '{}'".format(tracker.sender_id))
            if self.on_circuit_break:
                # call a registered callback
                self.on_circuit_break(tracker, dispatcher)

    def _parse_message(self, message):
        # for testing - you can short-cut the NLU part with a message
        # in the format _intent[entity1=val1,entity=val2]
        # parse_data is a dict of intent & entities
        if (message.text.startswith(INTENT_MESSAGE_PREFIX) or
                message.text.startswith("_")):
            if RegexInterpreter.is_using_deprecated_format(message.text):
                warnings.warn(
                        "Parsing messages with leading `_` is deprecated and "
                        "will be removed. Instead, prepend your intents with "
                        "`{0}`, e.g. `{0}mood_greet` "
                        "or `{0}restart`.".format(INTENT_MESSAGE_PREFIX))
            parse_data = RegexInterpreter().parse(message.text)
        else:
            if isinstance(self.interpreter, RasaMultiNLUHttpInterpreter):
                parse_data = self.interpreter.parse(message.text, message.output_channel.language)
            else:
                parse_data = self.interpreter.parse(message.text)

        logger.debug("Received user message '{}' with intent '{}' "
                     "and entities '{}'".format(message.text,
                                                parse_data["intent"],
                                                parse_data["entities"]))
        return parse_data


def start_rules_pulling_in_worker(rules_server, wait_time_between_pulls, agent):
    # type: (EndpointConfig, int, Agent) -> None

    worker = Thread(target=_run_rules_pulling_worker,
                    args=(rules_server, wait_time_between_pulls, agent))
    worker.setDaemon(True)
    worker.start()


def _run_rules_pulling_worker(rules_server, wait_time_between_pulls, agent):
    # type: (EndpointConfig, int, Agent) -> None
    while True:
        try:
            agent.rules = SuperAgent.get_rules(rules_server)
        except Exception as e:
            logger.warning("Failed pulling rules from server: {}".format(e))
        finally:
            time.sleep(wait_time_between_pulls)



import glob
from jsonmerge import Merger
import yaml
import os
import io
from collections import OrderedDict


class DomainsMerger(object):
    def __init__(self, folder_path, prefix='domain', output='aggregated_domains.yaml'):
        self.output = os.path.join(folder_path, output)
        self.schema = {
            "properties": {
                "actions": {
                    "mergeStrategy": "arrayMergeById",
                    "mergeOptions": {"idRef": "/"}
                },
                "intents": {
                    "mergeStrategy": "arrayMergeById",
                    "mergeOptions": {"idRef": "/"}
                },
                "entities": {
                    "mergeStrategy": "arrayMergeById",
                    "mergeOptions": {"idRef": "/"}
                }
            }
        }
        path_pattern = u'{}/{}*.y*ml'.format(folder_path, prefix)
        self.list_of_files = glob.glob(path_pattern)
        self.jsons = []
        self.merger = Merger(self.schema)
        self.merged = {}
        yaml.add_representer(OrderedDict, DomainsMerger.represent_ordereddict)

    def _load(self):
        for f in self.list_of_files:
            with io.open(f, 'r', encoding='utf-8') as stream:
                self.jsons.append(yaml.load(stream))

    def merge(self):
        self._load()
        merger = Merger(self.schema)
        self.merged = self.jsons[0]
        for i in range(1, len(self.jsons), 1):
            self.merged = merger.merge(self.merged, self.jsons[i])
        return self

    @staticmethod
    def represent_ordereddict(dumper, data):
        value = []

        for item_key, item_value in data.items():
            node_key = dumper.represent_data(item_key)
            node_value = dumper.represent_data(item_value)

            value.append((node_key, node_value))

        return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

    yaml.add_representer(OrderedDict, represent_ordereddict)

    def dump(self):

        with io.open(self.output, 'w', encoding="utf-8") as outfile:
            yaml.dump(self.merged, outfile, default_flow_style=False, allow_unicode=True)




import io
import yaml
import re
from rasa_core.actions.action import Action
from rasa_core.events import ActionExecuted, ActionReverted


class InputValidator(object):
    def __init__(self, rules):
        self.rules = rules if rules is not None else []
        self.actions_to_ignore = ['action_listen', 'action_invalid_utterance']

    def ignore_action(self, action_name):
        self.actions_to_ignore.append(action_name)

    def find(self, after):
        for rule in self.rules:
            if rule.get('regex', True):
                if re.match(rule['after'], after):
                    return rule
            elif rule['after'] == after:
                return rule
        return None

    def get_error(self, parse_data, tracker):
        previous_action = self._get_previous_action(tracker)
        if previous_action is None:
            return None
        return self._get_error(parse_data, previous_action)

    def _get_error(self, parse_data, previous_action):
        rule = self.find(previous_action)
        if rule is None:
            return None
        for expected in rule['expected']:
            intent_ok = 'intents' not in expected or parse_data["intent"]["name"] in expected['intents']
            parse_entities = map(lambda e: e['entity'], parse_data['entities'])

            # ok if no entities are expected or if expected entities are a subset of parse_entities
            entities_ok = 'entities' not in expected or set(expected['entities']).issubset(parse_entities)
            if entities_ok and intent_ok:
                return None
        return rule['error_template']

    @staticmethod
    def _load_yaml(rules_file):
        with io.open(rules_file, 'r', encoding='utf-8') as stream:
            try:
                return yaml.load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(exc)

    def _get_previous_action(self, tracker):
        action_listen_found = False
        for i in range(len(tracker.events) - 1, -1, -1):
            if i == 0:
                return None
            if type(tracker.events[i]) is ActionExecuted \
                    and action_listen_found is False \
                    and tracker.events[i].action_name not in self.actions_to_ignore:
                return tracker.events[i].action_name

        return None


class ActionInvalidUtterance(Action):
    def __init__(self, template):
        self.template = template

    def run(self, dispatcher, tracker, domain):
        dispatcher.utter_template(self.template, tracker)

        # utter error message
        latest_bot_message = {"text": tracker.latest_bot_utterance.text}

        # repeat previous question
        if tracker.latest_bot_utterance.data is not None:
            for key in tracker.latest_bot_utterance.data.keys():
                latest_bot_message[key] = tracker.latest_bot_utterance.data[key]

        dispatcher.utter_response(latest_bot_message)
        return [ActionReverted(), ActionReverted()]

    def name(self):
        return 'action_invalid_utterance'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import re
import uuid
import glob
import random
import logging
import argparse

import requests

from rasa_core import utils
from rasa_core.nlg import TemplatedNaturalLanguageGenerator

logger = logging.getLogger(__name__)
R_TITLE = re.compile('##\s*(.+)')
R_USER = re.compile('\*\s*([^\[]+)(?:\s+(\[.+\]$))?')
R_ACTION = re.compile('\s+- (utter_\S+)(?:\s+(\[.+\]$))?')

class TestNLG(TemplatedNaturalLanguageGenerator):
    def generate(self, template_name, tracker, output_channel, **kwargs):
        return {"text": template_name}


class TestSession(object):
    def __init__(self,
                 host=None,
                 test_cases=None,
                 shuffle=False,
                 distinct=True,
                 ):

        self.endpoint = '{host}/webhooks/rest/webhook'.format(host=(host if host.startswith('http://') else 'http://' + host))
        self.distinct = distinct
        self.shuffle = shuffle
        self.stories = self._build_stories_from_path(test_cases)
        if shuffle:
            random.shuffle(self.stories)

    def _get_response(self, sender_id, message):
        response = requests.post(self.endpoint, json={
            "sender": sender_id,
            "message": message,
        })

        logger.debug('Response incoming with status code {}'.format(response.status_code))
        payload = response.json()

        if isinstance(payload, list):
            return list(map(lambda obj: obj['text'], payload))
        elif isinstance(payload, dict):
            return payload['text']
        else:
            raise TypeError('Rasa Core response in an unrecognized type: {type}\npayload = {payload}'.format(
                type=type(payload),
                payload=payload,
            ))

    def _run_story_test(self, sender_id, story):
        utils.print_color('\n## ' + story['title'].upper(), utils.bcolors.OKBLUE)
        self.failed = False
        logger.debug('Starting story: {story}'.format(story=story))
        for index, step in enumerate(story['steps']):
            if self.failed is True:
                return

            utterance = step.pop(0)

            logger.debug('Starting step {index}: {utterance}'.format(index=index, utterance=utterance))
            messages = self._get_response(sender_id, utterance)
            logger.debug('messages = {}'.format(messages))
            logger.debug('expected = {}'.format(step))
            utils.print_color('* {}'.format(utterance), utils.bcolors.OKBLUE)
            for message in messages:
                expected = step.pop(0) if len(step) > 0 else None
                if message == expected:
                    utils.print_color("  - {}".format(message), utils.bcolors.OKGREEN)
                else:
                    utils.print_color("  - {} (exp: {})".format(message, expected), utils.bcolors.FAIL)
                    utils.print_color("TEST CASE INTERRUPTED)".format(message, expected), utils.bcolors.FAIL)
                    self.failed = True

    @staticmethod
    def _concatenate_storyfiles(folder_path, prefix='test', output='aggregated_test_cases.md'):
        path_pattern = u'{}/{}*.md'.format(folder_path, prefix)
        filenames = glob.glob(path_pattern)
        with open(output, 'w') as outfile:
            for fname in filenames:
                with open(fname, 'r') as infile:
                    for line in infile:
                        outfile.write(line)
                    outfile.write("\n")

    def _should_append_line(self, query):
        logger.debug('Flags for this line: {}'.format(query))
        # check if there are flags at all
        if query is None:
            return True
        extract_flag = re.compile(r'^\[--(\w+)\]$')
        flags = [extract_flag.match(x).group(1) for x in query.split(' ')]
        if None in flags:
            logger.warning('Could not parse flags: {}\nThey should be in the format [--distinct] or [--shuffle]')

        distinct = 'distinct' in flags
        shuffle = 'shuffle' in flags

        logger.debug('Flags: {}'.format(flags))
        # Return false if either conditions not met
        if distinct:
            if not self.distinct:
                return False
        if shuffle:
            if not self.shuffle:
                return False

        # Otherwise return true
        return True

    def _build_stories_from_path(self, test_cases_path):
        logger.debug('BUILDING STORIES')

        test_cases_path = os.path.join(test_cases_path)
        if os.path.isdir(test_cases_path):
            output_path = os.path.join(test_cases_path, 'aggregated_tests_cases.md')
            TestSession._concatenate_storyfiles(test_cases_path, 'test', output_path)
            test_cases_path = output_path
        stories = []

        with open(test_cases_path, 'r') as test_file:
            lines = test_file.readlines()

        story = None
        for line in lines:
            m_story = R_TITLE.match(line)
            m_user = R_USER.match(line)
            m_action = R_ACTION.match(line)
            if m_story is not None:
                if story is not None:
                    stories.append(story)

                story = {
                    'title': m_story.group(1),
                    'steps': []
                }
            elif m_user is not None:
                if self._should_append_line(m_user.group(2)):
                    story['steps'].append(['/' + m_user.group(1)])
            elif m_action is not None:
                if self._should_append_line(m_action.group(2)):
                    story['steps'][-1].append(m_action.group(1))
        if story is not None:
            stories.append(story)
        return stories

    def run_test_cases(self):
        sender_id = str(uuid.uuid4())
        for story in self.stories:
            # self.agent.processor._get_tracker(sender_id)._reset()
            if self.distinct:
                sender_id = str(uuid.uuid4())
            self._run_story_test(sender_id, story)
        return True


def create_argparser():
    parser = argparse.ArgumentParser(
            description='runs the bot.')

    parser.add_argument('--host',
                        help="URL of the running rasa core instance",
                        required=True)
    parser.add_argument('-t', '--tests',
                        help="Test cases/stories",
                        required=True)
    parser.add_argument('-s', '--shuffle',
                        help="Shuffle",
                        action='store_true',
                        default=False)
    parser.add_argument('-u', '--distinct',
                        help="Unique id for each case",
                        action='store_true',
                        default=False)
    parser.add_argument('-v', '--verbose',
                        help="Show debug logs",
                        action='store_true',
                        default=False)
    return parser


if __name__ == "__main__":
    parser = create_argparser()
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.WARNING, format='%(message)s')

    runner = TestSession(args.host, args.tests, args.shuffle, args.distinct)
    runner.run_test_cases()



import json
import re
from schema import Schema, And, Optional
from rasa_core.actions.action import Action
from rasa_core.events import SlotSet


class Disambiguator(object):
    def __init__(self, disamb_rule=None, fallback_rule=None):

        self.disamb_schema = Schema({
            "trigger": And(str, len),
            "type": "suggest",
            Optional("max_suggestions", default=2): int,
            Optional("slot_name"): str,
            "display": {
                Optional("intro_template"): And(str, len),
                "text_template": And(str, len, error="text_template is required"),
                Optional("button_title_template_prefix", default="utter_disamb"): And(str, len),
                Optional("fallback_button"): {
                    "title": And(str, len, error="fallback button title is required"),
                    "payload": And(str, len, error="fallback button payload is required")
                },
                Optional("exclude", default=[]): list
            }
        })

        self.rephrase_schema = Schema({
            "trigger": And(str, len),
            "type": "rephrase",
            Optional("slot_name"): str,
            "display": {
                "rephrase_template": And(str, len),
                "yes_template": And(str, len),
                "no_template": And(str, len),
                "no_payload": And(str, len),
                Optional("exclude", default=[]): list,
            }
        })

        self.fallback_schema = Schema({
            "trigger": And(str, len),
            "display": {
                "text": And(str, len),
                Optional("buttons"):
                    [{"title": And(str, len, error="button title is required"),
                      "payload": And(str, len, error="button title is required")}]
            }
        })

        if disamb_rule:
            # avoid breaking change
            if 'type' not in disamb_rule:
                disamb_rule['type'] = 'suggest'

            if 'type' not in disamb_rule or disamb_rule['type'] not in ['suggest', 'rephrase']:
                raise ValueError('type must be either \'suggest\' or \'rephrase\'')

            if disamb_rule['type'] == 'rephrase':
                self.disamb_rule = self.rephrase_schema.validate(disamb_rule)
            else:
                self.disamb_rule = self.disamb_schema.validate(disamb_rule)
        else:
            self.disamb_rule = None

        self.fallback_rule = self.fallback_schema.validate(fallback_rule) if fallback_rule else None

    @staticmethod
    def is_triggered(parse_data, trigger):
        # pattern to match $0, $1, $2, ... and returning 0, 1, 2,... in match groups
        pattern = re.compile(r"\$(\d)")
        eval_string = trigger
        # matches: an array of intents indices to consider in intent_ranking
        matches = re.findall(pattern, trigger)
        for i in matches:
            # if not enough intents in ranking to apply the rule, policy rule can't be triggered
            if int(i) >= len(parse_data["intent_ranking"]):
                return False
            eval_string = re.sub(r'\$' + i, str(parse_data["intent_ranking"][int(i)]["confidence"]), eval_string)

        return len(parse_data["intent_ranking"]) and eval(eval_string, {'__builtins__': {}})

    def should_fallback(self, parse_data):
        if not self.fallback_rule or "intent_ranking" not in parse_data:
            return False

        return self.is_triggered(parse_data, self.fallback_rule["trigger"])

    def should_disambiguate(self, parse_data):
        if not self.disamb_rule or "intent_ranking" not in parse_data:
            return False

        return self.is_triggered(parse_data, self.disamb_rule["trigger"])

    def should_exclude(self, intent_name):
        return any([re.match(x, intent_name) for x in self.disamb_rule["display"]["exclude"]])

    def get_payloads(self, parse_data, intents, keep_entities=True):
        entities = ""
        if "entities" in parse_data and keep_entities:
            entities = json.dumps(dict(map(lambda e: (e["entity"], e["value"]), parse_data["entities"])))
        payloads = list(["/{}{}".format(i, entities) for i in intents])
        return payloads

    def get_intent_names(self, parse_data):
        intent_names = list(map(lambda x: x["name"], parse_data["intent_ranking"]))
        if self.disamb_rule["display"]["exclude"]:
            intent_names = list(filter(lambda x: not self.should_exclude(x), intent_names))
        if self.disamb_rule["type"] == "suggest":
            return intent_names[:self.disamb_rule["max_suggestions"]]
        else:
            return intent_names[:1]

    def disambiguate(self, parse_data, tracker, dispatcher, run_action):
        should_disambiguate = self.should_disambiguate(parse_data)

        if should_disambiguate:
            intents = self.get_intent_names(parse_data)
            self.disamb_rule['parse_data'] = parse_data
            action = ActionDisambiguate(self.disamb_rule, self.get_payloads(parse_data, intents), intents)
            run_action(action, tracker, dispatcher)
            return True

    def fallback(self, parse_data, tracker, dispatcher, run_action):
        should_fallback = self.should_fallback(parse_data)

        if should_fallback:
            self.fallback_rule['parse_data'] = parse_data
            action = ActionFallback(self.fallback_rule)
            run_action(action, tracker, dispatcher)
            return True


class ActionDisambiguate(Action):

    def __init__(self, rule, payloads, intents):
        self.rule = rule
        self.payloads = payloads
        self.intents = intents

    @staticmethod
    def get_disambiguation_message(dispatcher, rule, payloads, intents, tracker):

        def generate(template_name):
            print('test')
            templates = dispatcher.nlg.generate(template_name, tracker, dispatcher.output_channel)
            if isinstance(templates, dict):
                return templates["text"]
            elif isinstance(templates, list) and len(templates):
                return templates[0]["text"]

            raise TypeError('templates must be a dictionary or a non-empty list')

        if rule["type"] == 'suggest':
            buttons = list(
                [{"title": generate("{}_{}".format(rule["display"]["button_title_template_prefix"], i[0])),
                  "payload": i[1]} for i in zip(intents, payloads)])

            if "fallback_button" in rule["display"] and \
                    "title" in rule["display"]["fallback_button"] and \
                    "payload" in rule["display"]["fallback_button"]:
                buttons.append({
                    "title": generate(rule["display"]["fallback_button"]["title"]),
                    "payload": rule["display"]["fallback_button"]["payload"]
                })

            disambiguation_message = {
                "text": generate(rule["display"]["text_template"]),
                "buttons": buttons
            }


        else:
            disambiguation_message = {
                "text": generate("{}_{}".format(rule["display"]["rephrase_template"], intents[0])),
                "buttons": [
                    {"title": generate(rule["display"]["yes_template"]), "payload": payloads[0]},
                    {"title": generate(rule["display"]["no_template"]),
                     "payload": rule["display"]["no_payload"]}
                ]
            }

        return disambiguation_message

    def run(self, dispatcher, tracker, domain):
        if "intro_template" in self.rule["display"]:
            dispatcher.utter_template(self.rule["display"]["intro_template"], tracker)

        disambiguation_message = self.get_disambiguation_message(dispatcher, self.rule, self.payloads, self.intents,
                                                                 tracker)
        dispatcher.utter_response(disambiguation_message)

        return_value = []
        if self.rule['slot_name']:
            return_value.append(SlotSet(self.rule['slot_name'], self.rule['parse_data']))
        return return_value

    def name(self):
        return 'action_disambiguate'


class ActionFallback(Action):

    def __init__(self, rule):
        self.rule = rule

    @staticmethod
    def get_fallback_message(dispatcher, rule, tracker):
        fallback_message = {
            "text": dispatcher.nlg.generate(rule["display"]["text"], tracker, dispatcher.output_channel)["text"]
        }
        if "buttons" in rule["display"]:
            buttons = list(
                [{"title": dispatcher.nlg.generate(button["title"], tracker, dispatcher.output_channel)["text"],
                  "payload": button["payload"]} for button in rule["display"]["buttons"]])
            fallback_message["buttons"] = buttons
        return fallback_message

    def run(self, dispatcher, tracker, domain):
        fallback_message = self.get_fallback_message(dispatcher, self.rule, tracker)
        dispatcher.utter_response(fallback_message)

        return_value = []
        if self.rule['slot_name']:
            return_value.append(SlotSet(self.rule['slot_name'], self.rule['parse_data']))
        return return_value

    def name(self):
        return 'action_fallback'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from rasa_addons.superagent.disambiguator import ActionDisambiguate
from rasa_addons.superagent.rules import Rules
from rasa_core.events import UserUttered
from rasa_core.processor import MessageProcessor
from rasa_core.dispatcher import Dispatcher

import logging

logger = logging.getLogger(__name__)


class SuperMessageProcessor(MessageProcessor):
    def __init__(self,
                 interpreter,  # type: NaturalLanguageInterpreter
                 policy_ensemble,  # type: PolicyEnsemble
                 domain,  # type: Domain
                 tracker_store,  # type: TrackerStore
                 generator,  # type: NaturalLanguageGenerator
                 action_endpoint=None,  # type: Optional[EndpointConfig]
                 max_number_of_predictions=10,  # type: int
                 message_preprocessor=None,  # type: Optional[LambdaType]
                 on_circuit_break=None,  # type: Optional[LambdaType]
                 create_dispatcher=None,  # type: Optional[LambdaType]
                 rules_file=None  # type: Optional[str]
                 ):

        self.rules = Rules(rules_file) if rules_file is not None else None
        super(SuperMessageProcessor, self).__init__(
            interpreter,
            policy_ensemble,
            domain,
            tracker_store,
            generator,
            action_endpoint,
            max_number_of_predictions,
            message_preprocessor,
            on_circuit_break
        )

        self.create_dispatcher = create_dispatcher
        if self.create_dispatcher is None:
            self.create_dispatcher = lambda sender_id, output_channel, nlg: Dispatcher(sender_id, output_channel, nlg)

    def _handle_message_with_tracker(self, message, tracker):
        # type: (UserMessage, DialogueStateTracker) -> None

        if message.parse_data:
            parse_data = message.parse_data
        else:
            parse_data = self._parse_message(message)

        # rules section #
        if self._rule_interrupts(parse_data, tracker, message):
            return
        # rules section - end #

        # don't ever directly mutate the tracker
        # - instead pass its events to log
        tracker.update(UserUttered(message.text, parse_data["intent"],
                                   parse_data["entities"], parse_data))
        # store all entities as slots
        for e in self.domain.slots_for_entities(parse_data["entities"]):
            tracker.update(e)

        logger.debug("Logged UserUtterance - "
                     "tracker now has {} events".format(len(tracker.events)))

    def _rule_interrupts(self, parse_data, tracker, message):
        if self.rules is not None:
            dispatcher = self.create_dispatcher(message.sender_id, message.output_channel, self.nlg)
            return self.rules.interrupts(dispatcher, parse_data, tracker, self._run_action)

    def _predict_and_execute_next_action(self, message, tracker):
        # this will actually send the response to the user

        dispatcher = self.create_dispatcher(message.sender_id,
                                            message.output_channel,
                                            self.nlg)
        # keep taking actions decided by the policy until it chooses to 'listen'
        should_predict_another_action = True
        num_predicted_actions = 0

        self._log_slots(tracker)

        # action loop. predicts actions until we hit action listen
        while (should_predict_another_action
               and self._should_handle_message(tracker)
               and num_predicted_actions < self.max_number_of_predictions):
            # this actually just calls the policy's method by the same name
            action, policy, confidence = self.predict_next_action(tracker)

            should_predict_another_action = self._run_action(action,
                                                             tracker,
                                                             dispatcher,
                                                             policy,
                                                             confidence)
            num_predicted_actions += 1

        if (num_predicted_actions == self.max_number_of_predictions and
                should_predict_another_action):
            # circuit breaker was tripped
            logger.warn(
                "Circuit breaker tripped. Stopped predicting "
                "more actions for sender '{}'".format(tracker.sender_id))
            if self.on_circuit_break:
                # call a registered callback
                self.on_circuit_break(tracker, dispatcher)



import copy
import io
import logging
import re

import yaml
from rasa_core.events import ActionExecuted
from requests.exceptions import RequestException

from rasa_addons.disambiguation import Disambiguator
from rasa_addons.input_validation import InputValidator, ActionInvalidUtterance

logger = logging.getLogger(__name__)


class Rules(object):
    def __init__(self, rules_dict):
        self.rules_dict = {}
        self.input_validation = None
        self.allowed_entities = None
        self.intent_substitutions = None
        self.disambiguation_policy = None
        self.actions_to_ignore = ['action_listen', 'action_invalid_utterance']
        self.update(rules_dict)

    def update(self, rules_dict):
        self.rules_dict = rules_dict
        self.allowed_entities = rules_dict[
            "allowed_entities"] if rules_dict and "allowed_entities" in rules_dict else {}
        self.intent_substitutions = rules_dict[
            "intent_substitutions"] if rules_dict and "intent_substitutions" in rules_dict else []
        self.input_validation = InputValidator(
            rules_dict["input_validation"]) if rules_dict and "input_validation" in rules_dict else []
        self.disambiguation_policy = Disambiguator(
            rules_dict.get("disambiguation_policy", None) if rules_dict else None,
            rules_dict.get("fallback_policy", None) if rules_dict else None)

    def get(self):
        return self.rules_dict

    def interrupts(self, dispatcher, parse_data, tracker, run_action):

        # fallback has precedence
        if self.disambiguation_policy.fallback(parse_data, tracker, dispatcher, run_action) or \
                self.disambiguation_policy.disambiguate(parse_data, tracker, dispatcher, run_action):
            return True

        self.run_swap_intent_rules(parse_data, tracker)

        self.filter_entities(parse_data)

        if self.input_validation:
            error_template = self.input_validation.get_error(parse_data, tracker)
            if error_template is not None:
                self._utter_error_and_roll_back(dispatcher, tracker, error_template, run_action)
                return True

    @staticmethod
    def _utter_error_and_roll_back(dispatcher, tracker, template, run_action):
        action = ActionInvalidUtterance(template)
        run_action(action, tracker, dispatcher)

    def filter_entities(self, parse_data):

        if parse_data['intent']['name'] in self.allowed_entities.keys():
            filtered = list(filter(lambda ent: ent['entity'] in self.allowed_entities[parse_data['intent']['name']],
                                   parse_data['entities']))
        else:
            filtered = parse_data['entities']

        if len(filtered) < len(parse_data['entities']):
            # logging first
            logger.warn("entity(ies) were removed from parse stories")
            parse_data['entities'] = filtered

    def run_swap_intent_rules(self, parse_data, tracker):
        # don't do anything if no intent is present
        if parse_data["intent"]["name"] is None or parse_data["intent"]["name"] == "":
            return

        previous_action = self._get_previous_action(tracker)

        for rule in self.intent_substitutions:
            if Rules._swap_intent(parse_data, previous_action, rule):
                break

    @staticmethod
    def _swap_intent(parse_data, previous_action, rule):
        # don't do anything if no intent is present
        if parse_data["intent"]["name"] is None or parse_data["intent"]["name"] == "":
            return

        # for an after rule
        if previous_action and 'after' in rule and re.match(rule['after'], previous_action):
            return Rules._swap_intent_after(parse_data, rule)

        # for a general substitution
        elif 'after' not in rule and re.match(rule['intent'], parse_data['intent']['name']):
            return Rules.swap_intent_with(parse_data, rule)

    @staticmethod
    def _swap_intent_after(parse_data, rule):
        rule['unless'] = rule['unless'] if 'unless' in rule else []
        if parse_data['intent']['name'] not in rule['unless']:
            logger.warn(
                "intent '{}' was replaced with '{}'".format(parse_data['intent']['name'], rule['intent']))
            parse_data['intent']['name'] = rule['intent']
            parse_data.pop('intent_ranking', None)
            return True

    @staticmethod
    def swap_intent_with(parse_data, rule):

        def format(text, parse_data):
            return text.format(intent=parse_data["intent"]["name"])

        pd_copy = copy.deepcopy(parse_data)
        parse_data['intent']['name'] = rule['with']
        parse_data['intent_ranking'] = [{"name": rule['with'], "confidence": 1.0}]
        if 'entities' in rule and 'add' in rule["entities"]:
            for entity in rule["entities"]["add"]:
                if 'entities' not in parse_data:
                    parse_data['entities'] = []
                parse_data['entities'].append(
                    {"entity": format(entity["name"], pd_copy), "value": format(entity["value"], pd_copy)})
        return True

    def _get_previous_action(self, tracker):
        action_listen_found = False
        for i in range(len(tracker.events) - 1, -1, -1):
            if i == 0:
                return None
            if type(tracker.events[i]) is ActionExecuted \
                    and action_listen_found is False \
                    and tracker.events[i].action_name not in self.actions_to_ignore:
                return tracker.events[i].action_name

        return None

    @staticmethod
    def _load_yaml(rules_file):
        with io.open(rules_file, 'r', encoding='utf-8') as stream:
            try:
                return yaml.load(stream)
            except yaml.YAMLError as exc:
                raise ValueError(exc)

    @classmethod
    def load_from_remote(cls, endpoint):
        try:
            logger.debug("Requesting rules from server {}..."
                         "".format(endpoint.url))
            response = endpoint.request(method='get')

            if response.status_code in [204, 304]:
                logger.debug("Model server returned {} status code, indicating "
                             "that no new rules are available.".format(response.status_code))
                return None
            elif response.status_code == 404:
                logger.warning("Tried to fetch rules from server but got a 404 response")
                return None
            elif response.status_code != 200:
                logger.warning("Tried to fetch rules from server, but server response "
                               "status code is {}. We'll retry later..."
                               "".format(response.status_code))
            else:
                rules = response.json()
                return Rules(rules)

        except RequestException as e:
            logger.warning("Tried to fetch rules from server, but couldn't reach "
                           "server. We'll retry later... Error: {}."
                           "".format(e))

    @classmethod
    def load_from_file(cls, rules_file):
        try:
            logger.debug('Loading rules from the {} file'.format(rules_file))
            return Rules(Rules._load_yaml(rules_file))
        except Exception as e:
            raise e



import yaml
import io

def load_yaml(rules_file):
    with io.open(rules_file, 'r', encoding='utf-8') as stream:
        try:
            return yaml.load(stream)
        except yaml.YAMLError as exc:
            raise ValueError(exc)



from setuptools import setup
import os
import io

with io.open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
setup(
    name='rasa_addons',
    version='0.5.5',
    author='Nathan Zylbersztejn',
    description="Rasa Addons - Productivity tools for Rasa Core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['jsonmerge', 'pyyaml', 'schema', 'rasa_core>=0.11.0'],
    licence='Apache 2.0',
    url='https://mrbot.ai',
    author_email='human@mrbot.ai',
    packages=['rasa_addons']
)
